#!/usr/bin/env python
#
# Tests algorithms for convex hulls (2D, spherical, 3D)
# Delaunay triangulation, and Voronoi diagrams
#
# Args: random-number seed (any string), number of points, then
# List of opcode-operand pairs:
# power -- distance adjustment for the points -- >1: core, <1: shell
# offset -- in the first coordinate
# CH2D -- 2D convex-hull tests: output file - DXF 2D
# CH2Sph -- spherical convex-hull tests: output file - DXF 3D
# CH3D -- 3D convex-hull tests: output file - DXF 3D
# CH3Sph -- 4-hyperspherical convex-hull tests: output file - DXF 3D
# CH4D -- 4D convex-hull tests: output file is a dummy arg (no file generated)
# DT2D -- 2D Delaunay-triangulation tests: output file - DXF 2D
# DTSph -- spherical Delaunay-triangulation tests: output file - DXF 3D

import sys, math, random, time

from VectorAlgebra import *
from DXFWrite import *
import ConvexHull2D, ConvexHull3D, ConvexHullND, Delaunay

# Inputs
ranseed = sys.argv[1]
numpts = int(sys.argv[2])
pwradj = 1
offset = 0

def OpOpHandle(args):
	global pwradj
	global offset
	opcode = None
	
	for arg in args:
		if opcode == None:
			opcode = arg
		else:
			operand = arg
			
			if opcode == "power":
				pwradj = float(operand)
			elif opcode == "offset":
				offset = float(operand)
			elif opcode == "CH2D":
				TestConvexHull2D(operand)		
			elif opcode == "CH2Sph":
				TestConvexHull2Sph(operand)		
			elif opcode == "CH3D":
				TestConvexHull3D(operand)		
			elif opcode == "CH3Sph":
				TestConvexHull3Sph(operand)		
			elif opcode == "CH4D":
				TestConvexHull4D(operand)		
			elif opcode == "DT2D":
				TestDelaunayTriangulation2D(operand)		
			elif opcode == "DT2Sph":
				TestDelaunayTriangulation2Sph(operand)		
			elif opcode == "DT3D":
				TestDelaunayTriangulation3D(operand)		
			elif opcode == "DT3Sph":
				TestDelaunayTriangulation3Sph(operand)		
			
			opcode = None

# Random points

def PowerAdjust(vec):
	# Uses global variable pwradj
	vnrm = VectorNorm(vec)
	if vnrm == 0: return vec
	else: return VectorScalMult(vnrm**(pwradj-1),vec)

def DoOffset(vec):
	# Uses global variable offset
	# Does offset in the first dimension
	vlst = list(vec)
	vlst[0] += offset
	return vlst

def MakePoint(dim):
	vec = [random.gauss(0,1) for d in xrange(dim)]
	vec = PowerAdjust(vec)
	vec = DoOffset(vec)
	return vec

def MakePoints(dim):
	# Uses global variables ranseed, numpts
	random.seed(ranseed)
	ptlist = []
	for p in xrange(numpts):
		ptlist.append(MakePoint(dim))
	return ptlist

def MakePoints2D(): return MakePoints(2)
def MakePoints3D(): return MakePoints(3)
def MakePoints4D(): return MakePoints(4)

def MakePointsNSph(dim):
	random.seed(ranseed)
	sphvts = []
	while len(sphvts) < numpts:
		newsphvts = [NormalizeVector(MakePoint(dim)) \
			for p in xrange(numpts - len(sphvts))]
		newsphvts = [vt for vt in newsphvts if InnerProduct(vt,vt) > 0]
		sphvts += newsphvts
	return sphvts

def MakePoints2Sph(): return MakePointsNSph(3)
def MakePoints3Sph(): return MakePointsNSph(4)

# Mathematica-format dumps

StrType = type("")
ListType = type([])
TupleType = type(())

def MMADump(x):
	xtype = type(x)
	if xtype == ListType or xtype == TupleType:
		return '{%s}' % ",".join(map(MMADump,x))
	elif xtype == StrType:
		return '"%s"' % x
	else:
		return str(x)


# Tests

def TimeTesting(vts, NamedAlgs):
	NamedReses = []
	
	print "Number of vertices:", len(vts)
	
	for name, alg in NamedAlgs:
		t0 = time.clock()
		res = alg(vts)
		t1 = time.clock()
		print name, t1-t0
		NamedReses.append((name,res))
	
	print "Number of distinct solutions:", len(set([nr[1] for nr in NamedReses]))
	
	return NamedReses

def DT_Flips(dtres):
	if len(dtres) >= 4:
		return dtres[3]
	else:
		return ()

def DT_TimeTesting(vts, NamedAlgs):
	NamedReses = []
	
	print "Number of vertices:", len(vts)
	
	for name, alg in NamedAlgs:
		t0 = time.clock()
		res = alg(vts)
		t1 = time.clock()
		print name, t1-t0, DT_Flips(res)
		NamedReses.append((name,res))
	
	print "Number of distinct solutions:", len(set([nr[1][:2] for nr in NamedReses]))
	
	return NamedReses


# Convex hulls

CH2DAlgs = (
	("Graham Scan", ConvexHull2D.GrahamScan), 
	("Gift Wrapping", ConvexHull2D.GiftWrapping),
	("Divide and Conquer", ConvexHull2D.DivideConquer),
	("Chan Shattering", ConvexHull2D.ChanShattering),
	("QuickHull", ConvexHull2D.QuickHull),
	("Incremental", ConvexHull2D.Incremental),
	("Monotone Chain", ConvexHull2D.MonotoneChain)
	)

def TestConvexHull2D(outfile):
	vts = MakePoints2D()
	
	print "Convex Hull 2D Flat"
	namedhulls = TimeTesting(vts, CH2DAlgs)
	hull = namedhulls[0][1]
	print "Vertices in hull:", len(hull)
	print
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hull in namedhulls:
		writepolygon(f, [vts[v] for v in hull], "CH "+name, 1)
	
	writeend(f)
	f.close()


CH2SphAlgs = (("Sph Incremental", ConvexHull2D.SphIncremental),)

def TestConvexHull2Sph(outfile):
	vts = MakePoints2Sph()
	
	print "Convex Hull 2D Spherical (3-sphere)"
	namedhulls = TimeTesting(vts, CH2SphAlgs)
	hull = namedhulls[0][1]
	print "Vertices in hull:", len(hull)
	print
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,cardpoints,0.01,"Cardinal Points",0)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hull in namedhulls:
		writepolygonarc(f, [vts[v] for v in hull], "CH "+name, 1)
	
	writeend(f)
	f.close()
	

CH3DAlgs = (
	("Gift Wrapping", ConvexHull3D.GiftWrapping),
	("Divide and Conquer", ConvexHull3D.DivideConquer),
	("Chan Shattering", ConvexHull3D.ChanShattering),
	("QuickHull", ConvexHull3D.QuickHull),
	("Incremental", ConvexHull3D.Incremental),
	("ND Gift Wrapping", ConvexHullND.GiftWrapping),
	("ND Chan Shattering", ConvexHullND.ChanShattering),
	("ND QuickHull", ConvexHullND.QuickHull),
	("ND Incremental", ConvexHullND.Incremental),
	)

def TestConvexHull3D(outfile):
	vts = MakePoints3D()
	
	print "Convex Hull 3D Flat"
	namedhulls = TimeTesting(vts, CH3DAlgs)
	hull = namedhulls[0][1]
	print "Vertices in hull:", len(hull[0])
	print "Triangles in hull:", len(hull[1])
	print

	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hull in namedhulls:
		for tri in hull[1]:
			writepolygon(f, [vts[v] for v in tri], "CH "+name, 1)
	
	writeend(f)
	f.close()


CH3SphAlgs = (("Sph Incremental", ConvexHullND.SphIncremental),)

def TestConvexHull3Sph(outfile):
	vts = MakePoints3Sph()
	
	print "Convex Hull 3D Spherical (4-sphere)"
	namedhulls = TimeTesting(vts, CH3SphAlgs)
	hull = namedhulls[0][1]
	print "Vertices in hull:", len(hull[0])
	print "Triangles in hull:", len(hull[1])
	print
	
	# Need to go from 4D to 3D to get something displayable
	# One could do the 4th dimension as continuous color,
	# but DXF does not support that in a simple way
	rdvts = [v[1:] for v in vts]
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,rdvts,0.01,"Vertices",2)
	
	for name, hull in namedhulls:
		for tri in hull[1]:
			writepolygon(f, [rdvts[v] for v in tri], "CH "+name, 1)
	
	writeend(f)
	f.close()


CH4DAlgs = (
	("Gift Wrapping", ConvexHullND.GiftWrapping),
	("Chan Shattering", ConvexHullND.ChanShattering),
	("QuickHull", ConvexHullND.QuickHull),
	("Incremental", ConvexHullND.Incremental),
	)

def TestConvexHull4D(outfile):
	vts = MakePoints4D()
	
	print "Convex Hull 4D Flat"
	namedhulls = TimeTesting(vts, CH4DAlgs)
	hull = namedhulls[0][1]
	print "Vertices in hull:", len(hull[0])
	print "Tetrahedra in hull:", len(hull[1])
	print

	# No output file


def DistinctVertices(vtsets):
	return list(sorted(reduce(lambda a,b: a | b, map(set, vtsets))))

DT2DAlgs = (
	("Divide and Conquer", Delaunay.DivideConquer),
	("Incr Flip", Delaunay.IncrFlip),
	("Incr Flip Lin", Delaunay.IncrFlipLin),
	("Incr Flip Rad", Delaunay.IncrFlipRad),
	("Convex Hull", Delaunay.ConvexHull),
	)

def TestDelaunayTriangulation2D(outfile):
	vts = MakePoints2D()
	
	print "Delaunay Tessellation 2D Flat"
	namedhldts = DT_TimeTesting(vts, DT2DAlgs)
	hldt = namedhldts[0][1]
	hull = hldt[0][0]
	tris = hldt[1]
	print "Vertices in hull:", len(hull)
	print "Vertices in triangles:", len(DistinctVertices(tris))
	print "Triangles:", len(tris)
	print
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hldt in namedhldts:
		hull = hldt[0][0]
		tris = hldt[1]
		for tri in tris:
			writepolygon(f, [vts[v] for v in tri], "DT Tris "+name, 1)
		writepolygon(f, [vts[v] for v in hull], "DT Hull "+name, 2)
	
	writeend(f)
	f.close()

DTSphAlgs = (
	("Sph Incr Flip", Delaunay.SphIncrFlip),
	("Sph Incr Flip Lin", Delaunay.SphIncrFlipLin),
	("Sph Incr Flip Rad", Delaunay.SphIncrFlipRad),
	("Sph Convex Hull", Delaunay.SphConvexHull),
	)

def TestDelaunayTriangulation2Sph(outfile):
	vts = MakePoints2Sph()
	
	print "Delaunay Tesselation 2D Spherical (3-sphere)"
	namedhldts = DT_TimeTesting(vts, DTSphAlgs)
	hldt = namedhldts[0][1]
	hull = hldt[0][0]
	tris = hldt[1]
	print "Vertices in hull:", len(hull)
	print "Vertices in triangles:", len(DistinctVertices(tris))
	print "Triangles:", len(tris)
	print
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,cardpoints,0.01,"Cardinal Points",0)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hldt in namedhldts:
		hull = hldt[0][0]
		tris = hldt[1]
		for tri in tris:
			writepolygonarc(f, [vts[v] for v in tri], "DT Tris "+name, 1)
		writepolygonarc(f, [vts[v] for v in hull], "DT Hull "+name, 2)
	
	writeend(f)
	f.close()

DT3DAlgs = (
	("Convex Hull", Delaunay.ConvexHull),
	)

def TestDelaunayTriangulation3D(outfile):
	vts = MakePoints3D()
	
	print "Delaunay Tesselation 3D Flat"
	namedhldts = DT_TimeTesting(vts, DT3DAlgs)
	hldt = namedhldts[0][1]
	print "Boundary vertices:", len(hldt[0][0])
	print "Boundary triangles:", len(hldt[0][1])
	print "Tetrahedra vertices:", len(DistinctVertices(hldt[1]))
	print "Tetrahedra in tessellation:", len(hldt[1])
	print

	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,vts,0.01,"Vertices",2)
	
	for name, hldt in namedhldts:
		for tet in hldt[1]:
			for tri in SimplexSubs(tet):
				writepolygon(f, [vts[v] for v in tri], "DT Tets "+name, 1)
		for tri in hldt[0][1]:
			writepolygon(f, [vts[v] for v in tri], "DT Hull "+name, 2)
	
	writeend(f)
	f.close()

DT3SphAlgs = (
	("Convex Hull", Delaunay.SphConvexHull),
	)

def TestDelaunayTriangulation3Sph(outfile):
	vts = MakePoints3Sph()
	
	print "Delaunay Tesselation 3D Spherical (4-sphere)"
	namedhldts = DT_TimeTesting(vts, DT3SphAlgs)
	hldt = namedhldts[0][1]
	print "Boundary vertices:", len(hldt[0][0])
	print "Boundary triangles:", len(hldt[0][1])
	print "Tetrahedra vertices:", len(DistinctVertices(hldt[1]))
	print "Tetrahedra in tessellation:", len(hldt[1])
	print
		
	# Need to go from 4D to 3D to get something displayable
	# One could do the 4th dimension as continuous color,
	# but DXF does not support that in a simple way
	rdvts = [v[1:] for v in vts]
	
	f = open(outfile,"w")
	writebegin(f)
	
	writecrosspoints(f,rdvts,0.01,"Vertices",2)
	
	for name, hldt in namedhldts:
		for tet in hldt[1]:
			for tri in SimplexSubs(tet):
				writepolygon(f, [rdvts[v] for v in tri], "DT Tets "+name, 1)
		for tri in hldt[0][1]:
			writepolygon(f, [rdvts[v] for v in tri], "DT Hull "+name, 2)
	
	writeend(f)
	f.close()


# Handle the opcode-operand pairs
OpOpHandle(sys.argv[3:])
