#!/usr/bin/env python
#
# Convex hull 2D:
# plane, sphere
#
# Arg: list of vertices
# Returns: tuple of convex-hull vertices
# in counterclockwise order
#
# Flat
#
# GrahamScan
# GiftWrapping
# DivideConquer
# ChanShattering
# QuickHull
# Incremental
# MonotoneChain
#
# ChanShattering has an optional additional argument:
# which algorithm to use for the parts (default: GrahamScan)
#
# Spherical (3D directions)
#
# Incremental

from VectorAlgebra import *

# Output in canonical form

def Output(ixs): return tuple(RotateToMin(ixs))


# The Graham Scan

def GSSortFunc(iv1,iv2):
	c = ScalarCross(iv1[1],iv2[1])
	return - cmp(c,0)

def GrahamScan(vts):
	if len(vts) == 0: return ()
	
	# First point
	ix0 = IndexMin([vt[0] for vt in vts])
	
	# All the others
	vt0 = vts[ix0]
	ixvts = [(k, VectorSub(vt,vt0)) for k,vt in enumerate(vts) if k != ix0]
	ixvts.sort(GSSortFunc)
	hull = [ix0] + [iv[0] for iv in ixvts]
	
	# Remove concavities with the three-penny algorithm
	while True:
		ptkeep = len(hull)*[True]
		ncnc = 0
		for k in xrange(1,len(hull)-1):
			if not ptkeep[k-1]: continue
			if not TriOrient2D([vts[hull[ix]] for ix in (k-1,k,k+1)]):
				ncnc += 1
				ptkeep[k] = False
		if ncnc == 0: break
		hull = SelWithList(ptkeep, hull)
	
	return Output(hull)


# Gift Wrapping

def PartialGiftWrapping(vts, indef=True, nmax=0):
	# Degenerate cases
	if len(vts) <= 1:
		return (True, range(len(vts)))

	# First point
	ix0 = IndexMin([vt[0] for vt in vts])
	ixprev = None
	ixcurr = ix0
	
	# All the others
	iscomplete = False
	hull = [ix0]
	
	n = 0
	while indef or (n < nmax):
		n += 1
		
		vtcurr = vts[ixcurr]
		if ixprev == None:
			prevdir = (0,-1)
		else:
			prevdir = VectorSub(vtcurr,vts[ixprev])
		ixnext = None
		csamax = None
		for ix,vt in enumerate(vts):
			if ix == ixcurr: continue
			vtdir = NormalizeVector(VectorSub(vt,vtcurr))
			csa = InnerProduct(prevdir,vtdir)
			if ixnext == None:
				ixnext = ix
				csamax = csa
			elif csa > csamax:
				ixnext = ix
				csamax = csa
		if ixnext == ix0:
			iscomplete = True
			break
		hull.append(ixnext)
		ixprev = ixcurr
		ixcurr = ixnext
	
	return (iscomplete, hull)

def GiftWrapping(vts):
	iscomplete, hull = PartialGiftWrapping(vts)
	return Output(hull)


# Divide and Conquer

# Goes in increasing order on the first hull and decreasing order on the second hull

def TwoHullTangent(hlvts1, hlvts2, ix10, ix20):
	ix1 = ix10; ix2 = ix20
	hlen1 = len(hlvts1)
	hlen2 = len(hlvts2)
	
	while True:
		numadvs = 0
		
		vt0 = hlvts2[ix2]
		vt1 = hlvts1[ix1]
		ixnx = ix1 + 1
		if ixnx >= hlen1: ixnx = 0
		vt2 = hlvts1[ixnx]
		if TriOrient2D((vt0,vt2,vt1)):
			ix1 = ixnx
			numadvs += 1
		
		vt0 = hlvts1[ix1]
		vt1 = hlvts2[ix2]
		ixnx = ix2 - 1
		if ixnx < 0: ixnx = hlen2 - 1
		vt2 = hlvts2[ixnx]
		if TriOrient2D((vt0,vt1,vt2)):
			ix2 = ixnx
			numadvs += 1
		if numadvs == 0: break
	
	return (ix1,ix2)

# Recursive

def DivideConquer(vts):
	# For at most 3 points, find the hull by hand
	vlen = len(vts)
	if vlen <= 3:
		hull = range(vlen)
		if vlen == 3:
			if not TriOrient2D([vts[h] for h in hull]):
				hull = [0,2,1]
		return Output(hull)
	
	# Find the hulls of the two parts
	ixvts = list(enumerate(vts))
	ixvts.sort(lambda a,b: cmp(a[1][0],b[1][0]))
	nspl = (vlen+1)//2
	ixvts1 = ixvts[:nspl]
	ixvts2 = ixvts[nspl:]
	ixs1 = [iv[0] for iv in ixvts1]
	ixs2 = [iv[0] for iv in ixvts2]
	vts1 = [iv[1] for iv in ixvts1]
	vts2 = [iv[1] for iv in ixvts2]
	hull1 = DivideConquer(vts1)
	hull2 = DivideConquer(vts2)
	#hull1 = DivideConquer([(v[1],-v[0]) for v in vts1])
	#hull2 = DivideConquer([(v[1],-v[0]) for v in vts2])
	
	# Find the closest points
	for k,h in enumerate(hull1):
		if h == len(ixs1)-1:
			ixnr1 = k
			break
	for k,h in enumerate(hull2):
		if h == 0:
			ixnr2 = k
			break
	
	# Get the hulls into the input points' indexing
	hull1 = [ixs1[h] for h in hull1]
	hull2 = [ixs2[h] for h in hull2]
	
	# Find tangent lines between the hulls
	vtsh1 = [vts[h] for h in hull1]
	vtsh2 = [vts[h] for h in hull2]
	ix11, ix12 = TwoHullTangent(vtsh1, vtsh2, ixnr1, ixnr2)
	ix22, ix21 = TwoHullTangent(vtsh2, vtsh1, ixnr2, ixnr1)
	
	# Stitch the hulls together
	# If the top and bottom tangents meet at the same point,
	# then the hull segment has zero size
	if ix21 >= ix11:
		hull1 = hull1[ix11:ix21+1]
	else:
		hull1 = hull1[ix11:] + hull1[:ix21+1]
	if ix12 >= ix22:
		hull2 = hull2[ix22:ix12+1]
	else:
		hull2 = hull2[ix22:] + hull2[:ix12+1]
	
	return Output(hull1 + hull2)


# Chan Shattering

# Need partial gift wrapping for it

def RoundRatio(num,den):
	return int(float(num)/float(den)+0.5)

def GenChanShattering(vts, HullVertices, CHParts, PartGiftWrap, MapVertices):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 2:
		return CHParts(vts)
	
	# Loop through trial hull lengths
	nhtrial = 3
	while True:
		nregs = RoundRatio(nvts,nhtrial)
		segs = [RoundRatio(nvts*k,nregs) for k in xrange(nregs+1)]
		shathulls = []
		for k in xrange(nregs):
			hull = CHParts(vts[segs[k]:segs[k+1]])
			shathulls += VectorScalAdd(segs[k], HullVertices(hull))
		if nregs == 1:
			break
		iscomplete, hull = \
			PartGiftWrap([vts[h] for h in shathulls], False, nhtrial)
		if iscomplete:
			hull = MapVertices(shathulls, hull)
			break
		
		nhtrial *= 3

	return hull

def CSHullVertices(hull): return hull

def CSMapVertices(shs,hull): return Output([shs[h] for h in hull])

def ChanShattering(vts, CHParts=GrahamScan):
	return GenChanShattering(vts, CSHullVertices, \
		CHParts, PartialGiftWrapping, CSMapVertices)


# QuickHull

# Returns indices of positive-side vertices
# and the maximum-distance one
def QHExternalVertices(vts, vtixs, seg):
	vt1 = vts[seg[0]]
	vt2 = vts[seg[1]]
	tdval = TriAreaData2D((vt2,vt1))
	ixvts = [(ix, TriAreaEval2D(tdval, vts[ix])) for ix in vtixs]
	ixpos = [iv[0] for iv in ixvts if iv[1] > 0]
	if len(ixpos) > 0:
		ixmax = MaxBySortByFunc(ixvts, lambda a: a[1])[0]
	else:
		ixmax = None
	return (ixpos, ixmax)


def QuickHull(vts):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 2:
		return tuple(range(nvts))
	
	# Starting points: first two
	ix1 = IndexMin([vt[0] for vt in vts])	
	ix2 = IndexMax([vt[0] for vt in vts])
	
	# Starting points: third one
	ixvts = [ix for ix in xrange(nvts) if ix not in (ix1,ix2)]
	ixvtria = []
	vt1 = vts[ix1]
	vt2 = vts[ix2]
	tdval = TriAreaData2D((vt1,vt2))
	for iv in ixvts:
		ta = TriAreaEval2D(tdval, vts[iv])
		# ta = TriArea2D((vt1,vt2,vts[iv]))
		ixvtria.append((iv,ta))
	
	ixng, tang = MinBySortByFunc(ixvtria, lambda v: v[1])
	ixps, taps = MaxBySortByFunc(ixvtria, lambda v: v[1])
	tang *= -1
	if taps >= tang:
		ix3 = ixps
		hull = [ix1, ix2, ix3]
	else:
		ix3 = ixng
		hull = [ix2, ix1, ix3]
	ixvts = [ix for ix in ixvts if ix != ix3]
	
	# Find the outward vertices for each edge
	extvts = {}
	hlen = len(hull)
	for k in xrange(hlen):
		kn = (k+1) % hlen
		seg = (hull[k], hull[kn])
		extvts[seg] = QHExternalVertices(vts, ixvts, seg)
	
	while True:
		# Search for an edge with outward points
		kx = None
		kxn = None
		hlen = len(hull)
		for k in xrange(hlen):
			kn = (k+1) % hlen
			seg = (hull[k],hull[kn])
			evs, ixmax = extvts[seg]
			if len(evs) > 0:
				kx = k
				kxn = kn
				break
		
		if kx == None: break
		
		# Find a point on the hull from it
		ixh = hull[kx]
		ixhn = hull[kxn]
		evs = [ix for ix in evs if ix != ixmax]
		
		if kxn < kx:
			hull = hull[kxn:kx+1] + [ixmax]
		else:
			hull = hull[kxn:] + hull[:kx+1] + [ixmax]
		
		# Find the outward points for the new hull segments 
		seg1 = (ixh, ixmax)
		seg2 = (ixmax, ixhn)
		extvts[seg1] = QHExternalVertices(vts, evs, seg1)
		extvts[seg2] = QHExternalVertices(vts, evs, seg2)
	
	return Output(hull)


# Incremental

def GenIncremental(vts, tridatafunc, triorient):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 1: return tuple(range(nvts))
	
	tridata = {}
	hull = range(2)
	hlen = len(hull)
	for k in xrange(hlen):
		kn = (k+1) % hlen
		hlk = hull[k]
		hlkn = hull[kn]
		vt1 = vts[hlk]
		vt2 = vts[hlkn]
		tridata[(hlk,hlkn)] = tridatafunc((vt2,vt1))
		
	for ix in xrange(2,nvts):
		# Make triangles between the point and pairs of neighbors on the hull.
		# Do any of them have the polarity expected for a nearby outside point?
		hlen = len(hull)
		bgns = hlen*[False]
		ends = hlen*[False]
		IsOutside = False
		AllOutside = True
		vt0 = vts[ix]
		
		for k in xrange(hlen):
			kn = (k+1) % hlen
			hlk = hull[k]
			hlkn = hull[kn]
			tdval = tridata[(hlk,hlkn)]
			if triorient(tdval, vt0):
				bgns[k] = True
				ends[kn] = True
				IsOutside = True
			else:
				AllOutside = False
		
		if AllOutside:
			hull = []
			break
		
		if not IsOutside: continue
		
		for k in xrange(hlen):
			kp = (k-1) % hlen
			if bgns[k] and not bgns[kp]:
				ixbgn = k
				break
		
		for k in xrange(hlen):
			kn = (k+1) % hlen
			if ends[k] and not ends[kn]:
				ixend = k
				break
		
		# Remove the points that will become inside and add the new one to the hull.
		if ixbgn < ixend:
			hull = hull[ixend:] + hull[:ixbgn+1] + [ix]
		elif ixbgn > ixend:
			hull = hull[ixend:ixbgn+1] + [ix]
		else:
			break
		
		hlk = hull[-2]
		hlkn = hull[-1]
		vt1 = vts[hlk]
		vt2 = vts[hlkn]
		tridata[(hlk,hlkn)] = tridatafunc((vt2,vt1))
		
		hlk = hull[-1]
		hlkn = hull[0]
		vt1 = vts[hlk]
		vt2 = vts[hlkn]
		tridata[(hlk,hlkn)] = tridatafunc((vt2,vt1))
	
	return Output(hull)

def Incremental(vts):
	return GenIncremental(vts, TriAreaData2D, TriOrientEval2D)

def SphIncremental(vts):
	return GenIncremental(vts, TriAreaData2Sph, TriOrientEval2Sph)


# Monotone Chain

# Each individual segment, either upper or lower.
# Gets vertices, indices in order - either upward or downward

def MonotoneChainSegment(vts, ixs):
	hseg = ixs[:1]
	
	for ix in ixs[1:]:
		vt0 = vts[ix]
		hsend = len(hseg)
		while hsend >= 2:
			vt1 = vts[hseg[hsend-1]]
			vt2 = vts[hseg[hsend-2]]
			if not TriOrient2D((vt0,vt1,vt2)): break
			hsend -= 1
		hseg = hseg[:hsend] + [ix]
	
	return hseg

def MonotoneChain(vts):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 1: return tuple(range(nvts))
	
	# Increasing and decreasing order
	ixvts = list(enumerate(vts))
	ixvts.sort(lambda a,b: cmp(a[1],b[1]))
	ixs1 = [iv[0] for iv in ixvts]
	ixs2 = list(reversed(ixs1))
	
	# The two hull halves
	hull1 = MonotoneChainSegment(vts, ixs1)
	hull2 = MonotoneChainSegment(vts, ixs2)
	hull = hull1[:-1] + hull2[:-1]
	return Output(hull)
