#!/usr/bin/env python
#
# Convex hull 3D:
# 3D flat space
# Surface of 4D hypersphere
#
# Arg: list of vertices
# Returns: (convex-hull vertices, convex-hull triangles)
# Specifying both because 1 or 2 vertices have vertices without triangles
#
# GiftWrapping
# DivideConquer
# ChanShattering
# QuickHull
# Incremental
#
# ChanShattering has an optional additional argument:
# which algorithm to use for the parts (default: GiftWrapping)
#
# Hypersphere (4D directions):
#
# SphIncremental

from VectorAlgebra import *
import ConvexHull2D

# 3D convex-hull output format:
# ((list of vertices in increasing order), (list of triangles)).
# This function composes the output from a list of triangles.

def OutputFromTriangles(hulltris):
	hullvts = reduce(lambda a,b: a | b, map(set,hulltris))
	hullvts = tuple(sorted(hullvts))
	
	sortedhulltris = [tuple(RotateToMin(tri)) for tri in hulltris]
	sortedhulltris = tuple(sorted(sortedhulltris))
	return (hullvts, sortedhulltris)


# Gift Wrapping

def GWMakeProjection(vts):
	refpt = VectorMean(vts[:2])
	diffvec = VectorSub(vts[1], vts[0])
	dfcovec = VectorScalMult(1.0/VectorAbsSqr(diffvec), diffvec)
	return (refpt, diffvec, dfcovec)

def GWPartProj(prj, vtdir):
	pvdr = InnerProduct(prj[2], vtdir)
	return VectorSub(vtdir, VectorScalMult(pvdr,prj[1]))

def GWProjectOut(prj, vt):
	vtdiff = VectorSub(vt, prj[0])
	return GWPartProj(prj, vtdiff)

def PartialGiftWrapping(vts, indef=True, nmax=0):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 2:
		return (True, (tuple(range(nvts)), ()))
	
	# Find a farthest-out point
	ix0 = IndexMin([vt[0] for vt in vts])
	
	# Find the closest-to-outward point using its direction
	vt = vts[ix0]
	ixvts = AddIndices(vts)
	ixvts1 = [(iv[0], NormalizeVector(VectorSub(iv[1],vt))[0])
		for iv in ixvts if iv[0] != ix0]
	ix1 = MinBySortByFunc(ixvts1, lambda a: a[1])[0]
	
	# Find the triangles
	bdryedges = []
	bdryoppoverts = []
	iscomplete = False
	initial = True;
	hulltris = []
	
	n = 0
	while indef or (n < nmax):
		n += 1
		
		# Find search direction
		if initial:
			seg = (ix0,ix1)
			prj = GWMakeProjection([vts[s] for s in seg])
			prevdir = CrossProduct((-1,0,0), prj[1])
			initial = False
		else:
			seg = bdryedges[0]
			ix = bdryoppoverts[0]
			prj = GWMakeProjection([vts[s] for s in seg])
			prevdir = GWProjectOut(prj, vts[ix])
			VectorScalMultBy(-1, prevdir)
		
		# Select the vertex that's closest to that direction.
		ixvts1 = [(iv[0], \
			InnerProduct(prevdir, NormalizeVector(GWProjectOut(prj,iv[1])))) \
				for iv in ixvts if iv[0] not in seg]
		ix = MaxBySortByFunc(ixvts1, lambda a: a[1])[0]

		# Compose the new triangle from that vertex
		newtri = (seg[1], seg[0], ix)
		hulltris.append(newtri)
		
		triedixs = ((1,2),(2,0),(0,1))
		triedges = [tuple([newtri[ev] for ev in ed]) for ed in triedixs]
		trioppoverts = newtri
		
		# Add its edges to the list of edges, but subtract where they duplicate.
		bdrysel = len(bdryedges)*[True]
		trisel = 3*[True]
		for k in xrange(3):
			triseg = tuple(reversed(triedges[k]))
			for ix, bdseg in enumerate(bdryedges):
				if bdseg == triseg:
					bdrysel[ix] = False
					trisel[k] = False
		
		bdryedges = SelWithList(bdrysel,bdryedges) + \
			SelWithList(trisel,triedges)
		bdryoppoverts = SelWithList(bdrysel,bdryoppoverts) + \
			SelWithList(trisel,trioppoverts)
		
		if len(bdryedges) == 0:
			iscomplete = True
			break
	
	hull = OutputFromTriangles(hulltris)
	return (iscomplete, hull)

def GiftWrapping(vts): return PartialGiftWrapping(vts)[1]


# Divide and Conquer

def DCFindNeighbors(tris, vts, ngbrix):
	if len(tris) > 0:
		# Collect each vertex's neighbors for each triangle
		# Remove duplicates as one does so
		for tri in tris:
			for vt in tri:
				if vt not in ngbrix:
					ngbrix[vt] = set()
				for tv in tri:
					if tv != vt:
						ngbrix[vt].add(tv)
		
	else:
		# With no triangles, use the vertex list instead 
		vtset = set(vts)
		for vt in vts:
			ngbrix[vt] = vtset - set((vt,))
	
	for vt in ngbrix.keys():
		ngbrix[vt] = tuple(ngbrix[vt])

def DCTrimTriangles(origtris, rngeds):
	tris = map(tuple,origtris)
	if len(tris) == 0: return []
	if len(rngeds) == 0: return []
	
	# Get the triangle and edge references, their references to each other
	trieds = {}
	edtri = {}
	for tri in tris:
		TriEdgeRefs(tri, trieds, edtri)
	
	# Ring edges: don't expand past those
	eduse = {}
	for tri in tris:
		for ed in trieds[tri]:
			eduse[ed] = True
	for ed in rngeds:
		eduse[ed] = False
	
	# Find which existing triangles are accessible from the ring edges.
	# 0 = unchecked, 1 = checked
	rngtris = [edtri[ed] for ed in rngeds]
	tristat = {}
	for tri in rngtris:
		tristat[tri] = 0
	unchecked = True
	
	while unchecked:
		unchecked = False
		for k, tri in enumerate(rngtris):
			if tristat[tri] == 0:
				unchecked = True
				
				# Look at a triangle's neighboring triangles
				for ed in trieds[tri]:
					if eduse[ed]:
						trix = edtri[tuple(reversed(ed))]
						if trix not in tristat:
							rngtris.append(trix)
							tristat[trix] = 0
				
				tristat[tri] = 1
	
	return list(set(tris) - set(rngtris))

# Recursive

def DivideConquer(vts):
	# For at most 3 points, find the hull by hand
	nvts = len(vts)
	if nvts <= 3:
		hullvts = tuple(range(nvts))
		if nvts == 3:
			hulltris = ((0,1,2),(0,2,1))
		else:
			hulltris = ()
		hull = (hullvts, hulltris)
		return hull
	
	# Find the hulls of the two parts
	ixvts = AddIndices(vts)
	ixvts = list(sorted(ixvts, None, lambda a: a[1][0]))
	nspl = int((nvts+1)/2)
	ixvts1 = ixvts[:nspl]
	ixvts2 = ixvts[nspl:]
	ixs1 = [iv[0] for iv in ixvts1]
	ixs2 = [iv[0] for iv in ixvts2]
	vts1 = [iv[1] for iv in ixvts1]
	vts2 = [iv[1] for iv in ixvts2]
	hull1 = DivideConquer(vts1)
	hull2 = DivideConquer(vts2)
	#hull1 = DivideConquer([RotateLeft(v) for v in vts1])
	#hull2 = DivideConquer([RotateLeft(v) for v in vts2])
	#hull1 = DivideConquer([(v[1],v[2],v[0]) for v in vts1])
	#hull2 = DivideConquer([(v[1],v[2],v[0]) for v in vts2])
	hullixs1 = [ixs1[h] for h in hull1[0]]
	hullixs2 = [ixs2[h] for h in hull2[0]]
	hullvts1 = [vts1[h] for h in hull1[0]]
	hullvts2 = [vts2[h] for h in hull2[0]]
	hulltris1 = [[ixs1[v] for v in t] for t in hull1[1]]
	hulltris2 = [[ixs2[v] for v in t] for t in hull2[1]]
	ngbrix1 = {}
	ngbrix2 = {}
	DCFindNeighbors(hulltris1, hullixs1, ngbrix1)
	DCFindNeighbors(hulltris2, hullixs2, ngbrix2)
	
	# Find the projected hulls
	projhull1 = ConvexHull2D.DivideConquer([vt[:2] for vt in hullvts1])
	projhull2 = ConvexHull2D.DivideConquer([vt[:2] for vt in hullvts2])

	# Find the closest points
	for k, vt in enumerate(projhull1):
		if vt == len(hullvts1) - 1:
			ixnr1 = k
			break
	for k, vt in enumerate(projhull2):
		if vt == 0:
			ixnr2 = k
			break

	# Find a tangent line between the projected hulls
	projhullvts1 = [hullvts1[vt] for vt in projhull1]
	projhullvts2 = [hullvts2[vt] for vt in projhull2]
	twtres = ConvexHull2D.TwoHullTangent(projhullvts1, projhullvts2, \
		ixnr1, ixnr2)
	ix1 = hullixs1[projhull1[twtres[0]]]
	ix2 = hullixs2[projhull2[twtres[1]]]
	ixin1 = ix1; ixin2 = ix2
	
	# Find belt between the two hulls
	# blixvts1 = zip(map(lambda ix: (1,ix), hullixs1), hullvts1)
	# blixvts2 = zip(map(lambda ix: (2,ix), hullixs2), hullvts2)
	# blixvts = blixvts1 + blixvts2
	numblix = len(hullixs1)*len(hullixs2)
	seg = [ix1, ix2]
	initial = True
	
	belttris = []
	hrvts1 = []
	hrvts2 = []
	hreds1 = []
	hreds2 = []
	for p in xrange(numblix):
		prj = GWMakeProjection([vts[s] for s in seg])
		if initial:
			prevdir = GWPartProj(prj, (0,0,-1))
			initial = False
		else:
			prevdir = GWProjectOut(prj, vts[ixp])
			VectorScalMultBy(-1, prevdir)
		
		blix1 = ngbrix1[seg[0]]
		blix2 = ngbrix2[seg[1]]
		blixvts01 = [((1,b),vts[b]) for b in blix1]
		blixvts02 = [((2,b),vts[b]) for b in blix2]
		blixvts1 = blixvts01 + blixvts02
		# blixvts1 = filter(lambda b: b[0][1] not in seg, blixvts)
		blixvts1 = [(iv[0], \
			InnerProduct(prevdir, NormalizeVector(GWProjectOut(prj,iv[1])))) \
				for iv in blixvts1]
		blix = MaxBySortByFunc(blixvts1, lambda a: a[1])[0]
		ih = blix[0]
		ixn = blix[1]
		
		# Compose the new triangle from that vertex
		newtri = (seg[1], seg[0], ixn)
		belttris.append(newtri)
		if ih == 1:
			hrvts1.append(ixn)
			hreds1.append((newtri[1],newtri[2]))
		elif ih == 2:
			hrvts2.append(ixn)
			hreds2.append((newtri[2],newtri[0]))
		ixp = seg[ih-1]; seg[ih-1] = ixn
		if seg[0] == ixin1 and seg[1] == ixin2: break
	
	# Remove inward triangles for both hulls
	hulltris1 = DCTrimTriangles(hulltris1, hreds1)
	hulltris2 = DCTrimTriangles(hulltris2, hreds2)
	
	# Merge the remaining ones together 
	return OutputFromTriangles(hulltris1 + belttris + hulltris2)


# Chan Shattering

def CSHullVertices(hull): return hull[0]

def CSMapVertices(shs,hull): return \
	OutputFromTriangles([tuple(RotateToMin([shs[v] for v in tri])) for tri in hull[1]]) \
	if len(hull[1]) > 0 else \
	(tuple(sorted([shs[v] for v in hull[0]])), ())

def ChanShattering(vts, CHParts=GiftWrapping):
	return ConvexHull2D.GenChanShattering(vts, CSHullVertices, \
		CHParts, PartialGiftWrapping, CSMapVertices)


# QuickHull

# Finds the triangles visible from a new vertex
# and replaces them with ones containing the new vertex.
# Also returns old triangles and new triangles

def VisTrisFromVert(vts, origtris, ixtri, ix, trieds, edtri, tetorient):
	# Find which existing triangles are visible from the point.
	# 0 = unchecked, 1 = visible, 2 = invisible
	oldtris = [ixtri]
	tristat = {}
	tristat[ixtri] = 0
	unchecked = True
	
	while unchecked:
		unchecked = False
		for k, tri in enumerate(oldtris):
			if tristat[tri] == 0:
				unchecked = True
				
				# Look at a triangle's neighboring triangles
				for ed in trieds[tri]:
					trix = edtri[tuple(reversed(ed))]
					if trix not in tristat:
						oldtris.append(trix)
						tristat[trix] = 0
				
				tetvts = [vts[v] for v in (tri[1],tri[0],tri[2],ix)]
				tristat[tri] = 1 if tetorient(tetvts) else 2
	
	# Find the horizon edges, those between visible and invisible triangles,
	# and the new triangles
	newtris = []
	for tri in oldtris:
		if tristat[tri] == 1:
			for ed in trieds[tri]:
				trix = edtri[tuple(reversed(ed))]
				if tristat[trix] == 2:
					newtri = (ed[0],ed[1],ix)
					newtris.append(newtri)
	
	for tri in newtris:
		TriEdgeRefs(tri, trieds, edtri)
	
	# Remove the old triangles
	oldtris = [tri for tri in oldtris if tristat[tri] == 1]
	updtris = list((set(origtris) - set(oldtris)) | set(newtris))
	
	return (updtris, oldtris, newtris)

# Returns indices of positive-side vertices and the maximum-distance one

def QHExternalVertices(vts, vtixs, tri):
	tetvts = [vts[v] for v in (tri[1],tri[0],tri[2])]
	tdval = TetVolumeData3D(tetvts)
	ixvts = [(ix, TetVolumeEval3D(tdval,vts[ix])) for ix in vtixs]
	ixpos = [iv[0] for iv in ixvts if iv[1] > 0]
	if len(ixpos) > 0:
		ixmax = MaxBySortByFunc(ixvts, lambda a: a[1])[0]
	else:
		ixmax = None
	return (ixpos, ixmax)

def QuickHull(vts):
	# Degenerate cases 
	nvts = len(vts)
	if nvts <= 3:
		hullvts = tuple(range(nvts))
		if nvts == 3:
			hulltris = ((0,1,2),(0,2,1))
		else:
			hulltris = ()
		hull = (hullvts, hulltris)
		return hull
	
	# Starting points: first two
	ix1 = IndexMin([vt[0] for vt in vts])
	ix2 = IndexMax([vt[0] for vt in vts])
	
	# Starting points: third one
	ixvts = [iv for iv in xrange(nvts) if iv not in (ix1,ix2)]
	ixvtria = [(iv, TriArea3DSqr([vts[v] for v in (ix1,ix2,iv)])) \
		for iv in ixvts]
	ivtmx = MaxBySortByFunc(ixvtria, lambda a: a[1])
	ix3 = ivtmx[0]
	
	# Starting points: fourth one
	ixvts = [iv for iv in ixvts if iv != ix3]
	tdval = TetVolumeData3D([vts[v] for v in (ix1,ix2,ix3)])
	ixvtetv = [(iv, TetVolumeEval3D(tdval, vts[iv])) \
		for iv in ixvts]
	# ixvtetv = [(iv, TetVolume3D(tdval, [vts[v] for v in (ix1,ix2,ix3,iv)])) \
	#	for iv in ixvts]
	ivtng = MinBySortByFunc(ixvtetv, lambda a: a[1])
	ivtps = MaxBySortByFunc(ixvtetv, lambda a: a[1])
	ixng = ivtng[0]
	ixps = ivtps[0]
	tvng = - ivtng[1]
	tvps = ivtps[1]
	if tvps >= tvng:
		ix4 = ixps
		hulltris = [(ix1,ix2,ix3),(ix1,ix3,ix4),(ix3,ix2,ix4),(ix2,ix1,ix4)]
	else:
		ix4 = ixng
		hulltris = [(ix2,ix1,ix3),(ix3,ix1,ix4),(ix2,ix3,ix4),(ix1,ix2,ix4)]
	ixvts = [iv for iv in ixvts if iv != ix4]
	
	# Find the outward vertices for each triangle
	trieds = {}
	edtri = {}
	extvts = {}
	for tri in hulltris:
		TriEdgeRefs(tri, trieds, edtri)
		extvts[tri] = QHExternalVertices(vts, ixvts, tri)
	
	while True:
		# Search for a triangle with outward points
		kx = None
		for k, tri in enumerate(hulltris):
			evs, ixmax = extvts[tri]
			if len(evs) > 0:
				# Found a non-empty outside volume
				kx = k
				break
		if kx == None: break
		
		# Get updated hull triangles,
		# old triangles (removed), and new triangles (added)
		hulltris, oldtris, newtris = VisTrisFromVert(vts, hulltris, tri, ixmax, \
			trieds, edtri, TetOrient3D)
		
		# Find the outward points for the new triangles.
		# Start with all those for the old triangles.
		evs = set()
		for tri in oldtris:
			evs |= set(extvts[tri][0])
		evs = list(evs)
		evs = [iv for iv in evs if iv != ixmax]
		for tri in newtris:
			extvts[tri] = QHExternalVertices(vts, evs, tri)
	
	return OutputFromTriangles(hulltris)


# Incremental

def GenIncremental(vts, tetorient, tetdatafunc, tetdataorient):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 2:
		return (True, (tuple(range(nvts)), ()))
	
	hulltris = [(0,1,2), (0,2,1)]
	tetdata = {}
	trieds = {}
	edtri = {}
	for tri in hulltris:
		tetdata[tri] = tetdatafunc([vts[tri[iv]] for iv in (1,0,2)])
		TriEdgeRefs(tri, trieds, edtri)
	
	for ix in xrange(3,nvts):
		kx = None
		for k, tri in enumerate(hulltris):
			tdval = tetdata[tri]
			if tetdataorient(tdval, vts[ix]):
				kx = k
				break
		if kx == None: continue
		
		hulltris, oldtris, newtris = VisTrisFromVert(vts, hulltris, tri, ix, \
			trieds, edtri, tetorient)
		if len(hulltris) == 0: break
		for tri in newtris:
			tetdata[tri] = tetdatafunc([vts[tri[iv]] for iv in (1,0,2)])
	
	return OutputFromTriangles(hulltris)

def Incremental(vts):
	return GenIncremental(vts, TetOrient3D, TetVolumeData3D, TetOrientEval3D)

def SphIncremental(vts):
	return GenIncremental(vts, SphSimplexOrient, SphSimplexVolData, SphSimplexOrientEval)
