#!/usr/bin/env python
#
# Convex hull ND:
# N-D flat space
# Surface of (N+1)-D hypersphere
# N is at least 3 here
#
# Arg: list of vertices
# Returns: (convex-hull vertices, convex-hull simplexes)
# Both lists are sorted, and each simplex's vertices are sorted
# though the last two may be interchanged to make the sorting
# an even permutation
#
# Flat space:
# 
# GiftWrapping
# ChanShattering
# QuickHull
# Incremental
#
# ChanShattering has an optional additional argument:
# which algorithm to use for the parts (default: GiftWrapping)
#
# Hypersphere ((N+1)D directions):
#
# SphIncremental

from VectorAlgebra import *
import ConvexHull2D

# ND convex-hull output format:
# ((list of vertices in increasing order), (list of faces)).
# This function composes the output from a list of faces.

def OutputFromSimplexes(hullfcs):
	hullvts = set()
	for face in hullfcs:
		hullvts |= set(face)
	
	hull = (hullvts, hullfcs)
	hull = tuple([tuple(sorted(hd)) for hd in hull])
	return hull


# Gift Wrapping

def GWMakeProjection(vts):
	refpt = VectorMean(vts)
	diffs = SubFirstFromVecSet(vts)
	trdiffs = Transpose(diffs)
	
	id = IdentityMatrix(len(vts[0]))
	
	tdfinv = Inverse(MatrixMult(diffs,trdiffs))
	tdfx = MatrixMult(MatrixMult(trdiffs,tdfinv),diffs)
	
	pjmat = MatrixSub(id, tdfx)
	return (refpt, pjmat)

def GWPartProj(prj, vdir): return MatVecMult(prj[1],vdir)

def GWProjectOut(prj, vt): return GWPartProj(prj, VectorSub(vt,prj[0]))

def PartialGiftWrapping(vts, indef=True, nmax=0):
	if len(vts) == 0: return (False, ((), ()))
	nvts = len(vts)
	ndims = len(vts[0])
	if ndims <= 2: return (False, ((), ()))
	
	# Degenerate cases
	if nvts < ndims:
		return (True, (tuple(range(nvts)), ()))
	if not (indef or nmax > 0):
		return (False, ((), ()))
	
	# Find a farthest-out point
	ix0 = IndexMin([vt[0] for vt in vts])
	
	
	# Find the closest-to-outward point using its direction
	vt = vts[ix0]
	ixvts = AddIndices(vts)
	ixvts1 = [iv for iv in ixvts if iv[0] != ix0]
	ixvts11 = [(iv[0], NormalizeVector(VectorSub(iv[1],vt))[0])
		for iv in ixvts1]
	ix1 = MinBySortByFunc(ixvts11, lambda a: a[1])[0]
	ixvts1 = [iv for iv in ixvts1 if iv[0] != ix1]
	face = [ix0, ix1]
	
	# Find the next closest-to-outward ones
	for k in xrange(ndims-2):
		prj = GWMakeProjection([vts[fx] for fx in face])
		ixvts11 = [(iv[0], NormalizeVector(GWProjectOut(prj,iv[1]))[0]) \
			for iv in ixvts1]
		ix = MinBySortByFunc(ixvts11, lambda a: a[1])[0]
		ixvts1 = [iv for iv in ixvts1 if iv[0] != ix]
		face.append(ix)
	
	# Adjust the orientation of the initial face
	face.sort()
	face = tuple(face)
	if SuperCrossProduct(SubFirstFromVecSet([vts[fx] for fx in face]))[0] < 0:
		face = SimplexFlip(face)
	
	iscomplete = False
	initial = True
	hullfcs = []
	
	n = 0
	while indef or (n < nmax):
		n += 1
		
		if initial:
			# Setup for the initial face
			hullfcs.append(face)
			bdrysbfcs = SimplexSubs(face)
			bdryoppoverts = list(reversed(face))
			initial = False
			continue
		
		# Find the remaining faces
		sbfc = bdrysbfcs[0]
		ix = bdryoppoverts[0]
		prj = GWMakeProjection([vts[fx] for fx in sbfc])
		prevdir = GWProjectOut(prj, vts[ix])
		VectorScalMultBy(-1, prevdir)
		
		# Select the vertex that's closest to that direction.
		ixvts1 = [(iv[0], \
			InnerProduct(prevdir, NormalizeVector(GWProjectOut(prj,iv[1])))) \
			for iv in ixvts if iv[0] not in sbfc]
		ix = MaxBySortByFunc(ixvts1, lambda a: a[1])[0]
		
		# Compose the new face from that vertex
		face = SimplexCanon(list(SimplexFlip(sbfc)) + [ix])
		hullfcs.append(face)
		fcsubfcs = SimplexSubs(face)
		fcoppoverts = list(reversed(face))
		
		# Add its subfaces to the list of subfaces,
		# but subtract where they duplicate.
		bdrysel = len(bdrysbfcs)*[True]
		fcsel = len(face)*[True]
		for k in xrange(len(face)):
			sbfcsel = tuple(SimplexFlip(fcsubfcs[k]))
			for ix, bdsbfc in enumerate(bdrysbfcs):
				if bdsbfc == sbfcsel:
					bdrysel[ix] = False
					fcsel[k] = False
		
		bdrysbfcs = SelWithList(bdrysel,bdrysbfcs) + \
			SelWithList(fcsel,fcsubfcs)
		bdryoppoverts = SelWithList(bdrysel,bdryoppoverts) + \
			SelWithList(fcsel,fcoppoverts)
		
		if len(bdrysbfcs) == 0:
			iscomplete = True
			break
	
	hull = OutputFromSimplexes(hullfcs)
	return (iscomplete, hull)
			
def GiftWrapping(vts): return PartialGiftWrapping(vts)[1]
			

# Chan Shattering

def CSHullVertices(hull): return hull[0]

def CSMapVertices(shs,hull): return \
	OutputFromSimplexes([tuple(SimplexCanon([shs[v] for v in tri])) for tri in hull[1]]) \
	if len(hull[1]) > 0 else \
	(tuple(sorted([shs[v] for v in hull[0]])), ())

def ChanShattering(vts, CHParts=GiftWrapping):
	return ConvexHull2D.GenChanShattering(vts, CSHullVertices, \
		CHParts, PartialGiftWrapping, CSMapVertices)


# QuickHull

# Find the faces visible from the a new vertex
# and replace them with ones containing the new vertex.
# Also returns the old faces and the new faces.

def VisFacesFromVert(vts, origfaces, startface, ix, fctosbfcs, sbfctofcs, orient):
	# Find which existing faces are visible from the vertex.
	# 0 = unchecked, 1 = visible, 2 = invisible
	oldfaces = [startface]
	facestat = {}
	facestat[startface] = 0
	unchecked = True
	
	while unchecked:
		unchecked = False
		for k, face in enumerate(oldfaces):
			if facestat[face] == 0:
				unchecked = True
				
				# Look at a face's neighboring faces
				for sbfc in fctosbfcs[face]:
					facex = sbfctofcs[SimplexFlip(sbfc)]
					if facex not in facestat:
						oldfaces.append(facex)
						facestat[facex] = 0
				
				faceixs = list(SimplexFlip(face)) + [ix]
				facevts = [vts[v] for v in faceixs]
				facestat[face] = 1 if orient(facevts) else 2
		
	# Find the horizon subfaces, those between visible and invisible faces,
	# and the new faces
	newfaces = []
	for face in oldfaces:
		if facestat[face] == 1:
			for sbfc in fctosbfcs[face]:
				facex = sbfctofcs[SimplexFlip(sbfc)]
				if facestat[facex] == 2:
					newface = SimplexCanon(list(sbfc) + [ix])
					newfaces.append(newface)
	
	for face in newfaces:
		SimplexSubRefs(face, fctosbfcs, sbfctofcs)
	
	# Remove the old faces
	oldfaces = [face for face in oldfaces if facestat[face] == 1]
	updfaces = list((set(origfaces) - set(oldfaces)) | set(newfaces))
	
	return (updfaces, oldfaces, newfaces)

# Returns indices of positive-side vertices and the maximum-distance one

def QHExternalVertices(vts, vtixs, face):
	fcvts = [vts[v] for v in SimplexFlip(face)]
	tdval = SimplexVolData(fcvts)
	ixvts = [(ix, SimplexVolEval(tdval, vts[ix])) for ix in vtixs]
	ixpos = [iv[0] for iv in ixvts if iv[1] > 0]
	if len(ixpos) > 0:
		ixmax = MaxBySortByFunc(ixvts, lambda a: a[1])[0]
	else:
		ixmax = None
	return (ixpos, ixmax)

def QuickHull(vts):
	if len(vts) == 0: return ((), ())
	nvts = len(vts)
	ndims = len(vts[0])
	if ndims <= 2: return ((), ())
	
	# Degenerate cases 
	if nvts <= ndims:
		hullvts = tuple(range(nvts))
		if nvts == ndims:
			hulltris = (hullvts, SimplexFlip(hullvts))
		else:
			hulltris = ()
		hull = (hullvts, hulltris)
		return hull
	
	ixvts = range(nvts)
	ixinit = []
	
	# First two points
	vtstrim = [vt[0] for vt in vts]
	ix = IndexMin(vtstrim)
	ixinit.append(ix)
	ix = IndexMax(vtstrim)
	ixinit.append(ix)
	ixvts = [iv for iv in ixvts if iv not in ixinit]
	
	# Remaining points
	for k in xrange(2,ndims+1):
		vtstrim = [vt[:k] for vt in vts]
		ixvtarea = [(iv, abs(SimplexVolume([vtstrim[v] for v in (ixinit+[iv])]))) \
			for iv in ixvts]
		ix = MaxBySortByFunc(ixvtarea, lambda a: a[1])[0]
		ixinit.append(ix)
		ixvts = [iv for iv in ixvts if iv != ix]
		
	ixinit.sort()
	if not SimplexOrient([vts[v] for v in ixinit]):
		SimplexFlipInPlace(ixinit)
	
	hullfcs = list(SimplexSubs(ixinit))
	
	# Find the subfaces for each face
	fctosbfcs = {}
	sbfctofcs = {}
	extsfs = {}
	for face in hullfcs:
		SimplexSubRefs(face, fctosbfcs, sbfctofcs)
		extsfs[face] = QHExternalVertices(vts, ixvts, face)
	
	while True:
		# Search for a face with vertices outside of it
		kx = None
		for k, face in enumerate(hullfcs):
			evs, ixmax = extsfs[face]
			if len(evs) > 0:
				# Found a non-empty outside volume
				kx = k
				break
		if kx == None: break
		
		# Get updated hull faces,
		# old faces (removed), and new faces (added)
		hullfcs, oldfcs, newfcs = VisFacesFromVert(vts, hullfcs, face, ixmax, \
			fctosbfcs, sbfctofcs, SimplexOrient)
		
		evs = set()
		for face in oldfcs:
			evs |= set(extsfs[face][0])
		evs = list(evs)
		evs = [iv for iv in evs if iv != ixmax]
		for face in newfcs:
			extsfs[face] = QHExternalVertices(vts, evs, face)
	
	return OutputFromSimplexes(hullfcs)


# Incremental

def GenIncremental(vts, extradims, orient, datafunc, dataorient):
	if len(vts) == 0: return ((), ())
	nvts = len(vts)
	ndims = len(vts[0])
	ndims -= extradims
	if ndims <= 2: return ((), ())
	
	# Degenerate cases
	if nvts < ndims:
		return (True, (tuple(range(nvts)), ()))
	
	hullvts = tuple(range(ndims))
	hullfcs = [hullvts, SimplexFlip(hullvts)]
	facedata = {}
	fctosbfcs = {}
	sbfctofcs = {}
	for face in hullfcs:
		facedata[face] = datafunc([vts[v] for v in SimplexFlip(face)])
		SimplexSubRefs(face, fctosbfcs, sbfctofcs)
	
	for ix in xrange(ndims,nvts):
		kx = None
		for k, face in enumerate(hullfcs):
			fcval = facedata[face]
			if dataorient(fcval, vts[ix]):
				kx = k
				break
		if kx == None: continue
		
		hullfcs, oldfcs, newfcs = VisFacesFromVert(vts, hullfcs, face, ix, \
			fctosbfcs, sbfctofcs, orient)
		if len(hullfcs) == 0: break
		for face in newfcs:
			facedata[face] = datafunc([vts[v] for v in SimplexFlip(face)])

	return OutputFromSimplexes(hullfcs)

def Incremental(vts):
	return GenIncremental(vts, 0, \
		SimplexOrient, SimplexVolData, SimplexOrientEval)

def SphIncremental(vts):
	return GenIncremental(vts, 1, \
		SphSimplexOrient, SphSimplexVolData, SphSimplexOrientEval)
