#!/usr/bin/env python
#
# Delaunay triangulation / tessellation:
# plane, sphere
#
# Arg: list of vertices
# Returns: convex hull, tessellation,
# upper-face tessellation (convex hull) or empty,
# number of flips (incremental + flipping) or none
#
# All of them are 2D except for the convex-hull one
#
# DivideConquer
# IncrFlip
# IncrFlipLin
# IncrFlipRad
# ConvexHull
#
# These ones are for a spherical surface.
# All of them are for a 3-sphere (2D surface)
# except for the convex-hull one
#
# SphIncrFlip
# SphIncrFlipLin
# SphIncrFlipRad
# SphConvexHull
#
# The incremental-flipping (IncrFlip) algorithms
# add vertices, then flip edges to improve the triangulation
# The "lin" ones add the vertices from left to right
# The "rad" ones add the vertices from their centroid outward
#
# The convex-hull one uses convex hulls,
# and it can work in any dimension of points
#
# Convex-hull functions and their defaults:
#
# BoundaryCH2 = ConvexHull2D.QuickHull
# BoundaryCH3 = ConvexHull3D.QuickHull
# BoundaryCHN = ConvexHullND.QuickHull
# SphBoundaryCH2 = ConvexHull2D.SphIncremental
# SphBoundaryCHN = ConvexHullND.SphIncremental
# TessCH3 = ConvexHull3D.QuickHull
# TessCHN = ConvexHullND.QuickHull

from VectorAlgebra import *
import ConvexHull2D, ConvexHull3D, ConvexHullND

# 2D, Divide and Conquer

def DivideConquer(vts):
	# For at most 3 points, find the hull by hand
	vlen = len(vts)
	if vlen <= 3:
		hull = tuple(range(vlen))
		tris = ()
		if vlen == 3:
			if not TriOrient2D([vts[h] for h in hull]):
				hull = (0,2,1)
			tris = (hull,)
		return ((hull,), tris, ())
	
	# Find the hulls of the two parts
	ixvts = list(enumerate(vts))
	ixvts.sort(lambda a,b: cmp(a[1][0],b[1][0]))
	nspl = (vlen+1)//2
	ixvts1 = ixvts[:nspl]
	ixvts2 = ixvts[nspl:]
	ixs1 = [iv[0] for iv in ixvts1]
	ixs2 = [iv[0] for iv in ixvts2]
	vts1 = [iv[1] for iv in ixvts1]
	vts2 = [iv[1] for iv in ixvts2]
	#hldt1 = DivideConquer(vts1)
	#hldt2 = DivideConquer(vts2)
	# Rotate the points 90d for calculating the child hulls,
	# so as to alternate the directions of splitting
	hldt1 = DivideConquer([(v[1],-v[0]) for v in vts1])
	hldt2 = DivideConquer([(v[1],-v[0]) for v in vts2])
	hull1 = hldt1[0][0]
	hull2 = hldt2[0][0]
	tris1 = hldt1[1]
	tris2 = hldt2[1]
	
	# Find the closest points
	for k,h in enumerate(hull1):
		if h == len(ixs1)-1:
			ixnr1 = k
			break
	for k,h in enumerate(hull2):
		if h == 0:
			ixnr2 = k
			break
	
	# Get the hulls and triangles into the input points' indexing
	hull1 = [ixs1[h] for h in hull1]
	hull2 = [ixs2[h] for h in hull2]
	tris1 = [tuple(RotateToMin([ixs1[v] for v in tri])) for tri in tris1]
	tris2 = [tuple(RotateToMin([ixs2[v] for v in tri])) for tri in tris2]
	trieds = {}
	edtri = {}
	for tri in tris1:
		TriEdgeRefs(tri,trieds,edtri)
	for tri in tris2:
		TriEdgeRefs(tri,trieds,edtri)
	
	# Find tangent lines between the hulls
	vtsh1 = [vts[h] for h in hull1]
	vtsh2 = [vts[h] for h in hull2]
	ix11, ix12 = ConvexHull2D.TwoHullTangent(vtsh1, vtsh2, ixnr1, ixnr2)
	ix22, ix21 = ConvexHull2D.TwoHullTangent(vtsh2, vtsh1, ixnr2, ixnr1)
	
	# Find the inside hulls for making the inside triangles
	# If the top and bottom tangents meet at the same point,
	# then the hull segment starts and ends at that point
	if ix11 > ix21:
		hli1 = hull1[ix21:ix11+1]
	else:
		hli1 = hull1[ix21:] + hull1[:ix11+1]
	if ix22 > ix12:
		hli2 = hull2[ix12:ix22+1]
	else:
		hli2 = hull2[ix12:] + hull2[:ix22+1]
	
	# Stitch the hulls together
	# If the top and bottom tangents meet at the same point,
	# then the hull segment has zero size
	if ix21 >= ix11:
		hull1 = hull1[ix11:ix21+1]
	else:
		hull1 = hull1[ix11:] + hull1[:ix21+1]
	if ix12 >= ix22:
		hull2 = hull2[ix22:ix12+1]
	else:
		hull2 = hull2[ix22:] + hull2[:ix12+1]
	hull = tuple(RotateToMin(hull1 + hull2))

	# Do the merged triangles from bottom to top.
	# Make the internal hulls both top to bottom
	hli1.reverse()
	hiix1 = len(hli1) - 1
	hiix2 = len(hli2) - 1
	rmtris1 = []
	rmtris2 = []
	adtris = []
	
	while True:
		adv1 = hiix1 > 0
		adv2 = hiix2 > 0
		
		# Unable to advance further?
		if (not adv1) and (not adv2): break
		crix1 = hli1[hiix1]
		crix2 = hli2[hiix2]
		
		# Left-side advance?
		if adv1:
			hlad1 = []
			rmtri1 = []
			nxix1 = hli1[hiix1-1]
			if TriOrient2D([vts[v] for v in (crix1,crix2,nxix1)]):
				# Look inward from the edge to find a suitable next point
				while True:
					ccdt1 = CircumData([vts[v] for v in (crix1,crix2,nxix1)])
					seg = (crix1,nxix1)
					if seg not in edtri: break
					tri = edtri[seg]
					ix = [v for v in tri if v not in seg][0]
					if not TriOrient2D([vts[v] for v in (crix1,crix2,ix)]):
						break
					if CircumInside(ccdt1,vts[ix]):
						hlad1.append(ix)
						rmtri1.append(tri)
						nxix1 = ix
					else:
						break
			else:
				adv1 = False
		
		# Right-side advance?
		if adv2:
			hlad2 = []
			rmtri2 = []
			nxix2 = hli2[hiix2-1]
			if TriOrient2D([vts[v] for v in (crix1,crix2,nxix2)]):
				# Look inward from the edge to find a suitable next point
				while True:
					ccdt2 = CircumData([vts[v] for v in (crix1,crix2,nxix2)])
					seg = (nxix2,crix2)
					if seg not in edtri: break
					tri = edtri[seg]
					ix = [v for v in tri if v not in seg][0]
					if not TriOrient2D([vts[v] for v in (crix1,crix2,ix)]):
						break
					if CircumInside(ccdt2,vts[ix]):
						hlad2.append(ix)
						rmtri2.append(tri)
						nxix2 = ix
					else:
						break
			else:
				adv2 = False
		
		# Both-sides advance? Find out which one
		if adv1 and adv2:
			if not CircumInside(ccdt1, vts[nxix2]):
				adv2 = False
			elif not CircumInside(ccdt2, vts[nxix1]):
				adv1 = False
			else:
				break
		
		if adv1:
			# Advance on left side
			if len(hlad1) > 0:
				hli1 = hli1[:hiix1] + hlad1
				hiix1 = len(hli1) - 1
			else:
				hiix1 -= 1
			if len(rmtri1) > 0:
				rmtris1 += rmtri1
			tri = tuple(RotateToMin([crix1,crix2,nxix1]))
			adtris.append(tri)
		elif adv2:
			# Advance on right side
			if len(hlad2) > 0:
				hli2 = hli2[:hiix2] + hlad2
				hiix2 = len(hli2) - 1
			else:
				hiix2 -= 1
			if len(rmtri2) > 0:
				rmtris2 += rmtri2
			tri = tuple(RotateToMin([crix1,crix2,nxix2]))
			adtris.append(tri)
		else:
			break
	
	tris = list(set(tris1) - set(rmtris1)) + \
		list(set(tris2) - set(rmtris2)) + \
		adtris
	tris.sort()
	return ((hull,), tuple(tris), ())


# 2D, Spherical - Incremental, Flipping

# Assembles hull from segments

def AssembleHull(segs):
	if len(segs) == 0: return ()
	
	first = min(map(min,segs))
	next = {}
	for seg in segs:
		next[seg[0]] = seg[1]
	
	vt = first
	hull = []
	while True:
		hull.append(vt)
		vt = next[vt]
		if vt == first: break
	
	return tuple(hull)

# Returns convex hull, triangles

def GenInitialTriangulation(vts, trieds, edtri, orient, datafunc, dataorient):
	# Degenerate cases
	nvts = len(vts)
	if nvts <= 2:
		return (tuple(range(nvts)), ())
	
	# Initial triangle
	tri = (0,1,2)
	if not orient([vts[v] for v in tri]): tri = (0,2,1)
	tris = [tri]
	TriEdgeRefs(tri, trieds, edtri)
	sidetest = {}
	sidetest[tri] = \
		[datafunc([vts[v] for v in ed]) for ed in TriEdges(tri)]
	hullsegs = [tuple(reversed(ed)) for ed in TriEdges(tri)]
	
	for ix in xrange(3,nvts):
		# What external triangles contain this vertex
		vt = vts[ix]
		newtris = [tuple(RotateToMin((ed[0],ed[1],ix))) \
			for ed in hullsegs]
		newtris = [tri for tri in newtris if orient([vts[v] for v in tri])]
		
		if len(newtris) > 0:
			# On the outside
			for tri in newtris:
				TriEdgeRefs(tri, trieds, edtri)
				sidetest[tri] = \
					[datafunc([vts[v] for v in ed]) for ed in TriEdges(tri)]
			tris += newtris
			newhullsegs = []
			for tri in newtris:
				newhullsegs += [tuple(reversed(ed)) for ed in TriEdges(tri)]
			hullsegs += newhullsegs
			hullsegs = [ed for ed in hullsegs if ed not in edtri]
		
		else:
			# On the inside
			# Search in reverse to check most recent triangles first
			kx = None
			for k in xrange(len(tris)-1, -1, -1):
				tri = tris[k]
				if all((dataorient(data,vt) for data in sidetest[tri])):
					kx = k
					break
			
			if kx != None:
				# Inside a triangle: split it in three
				tri = tris[kx]
				newtris = [tuple(RotateToMin((ed[0],ed[1],ix))) \
					for ed in TriEdges(tri)]
				for tri in newtris:
					TriEdgeRefs(tri, trieds, edtri)
					sidetest[tri] = \
						[datafunc([vts[v] for v in ed]) for ed in TriEdges(tri)]
				tris[kx] = newtris[0]
				tris += newtris[1:]
	
	hullsegs = [tuple(reversed(seg)) for seg in hullsegs]
	return (AssembleHull(hullsegs), tris)
			
def GenEdgeFlipping(vts, origtris, trieds, edtri, orient, ccdata, ccinside):
	# Index the triangles
	tris = origtris
	triixs = {}
	for k, tri in enumerate(tris):
		triixs[tri] = k
	
	# Find the edges, index them, and mark them as to be examined.
	# Their vertices are in sorted order to resolve ambiguity
	edcnts = {}
	for tri in tris:
		for ed in TriEdges(tri):
			edsort = tuple(sorted(ed))
			if edsort not in edcnts: edcnts[edsort] = 0
			edcnts[edsort] += 1
	eds = edcnts.keys()
	
	neds = len(eds)
	edixs = {}
	for k, ed in enumerate(eds):
		edixs[ed] = k
	edtests = neds*[True]
	
	# Find which ones are interior edges
	isinted = neds*[False]
	for k, ed in enumerate(eds):
		isinted[k] = edcnts[ed] > 1
	
	# Do as many passes as necessary
	ccdtri = {}
	totflips = []
	for p in xrange(16*neds):
		numflips = 0
		for k in xrange(neds):
			if not isinted[k]: continue
			if not edtests[k]: continue
			
			# Get the edge and its triangles and their vertices
			ed = eds[k]
			tri1 = edtri[ed]
			tri2 = edtri[tuple(reversed(ed))]
			for v in tri1:
				if v not in ed: ix1 = v; break
			for v in tri2:
				if v not in ed: ix2 = v; break
			
			# Find the circumcircles of the triangles.
			# Are the opposite vertices of the other triangles in them?
			if tri1 not in ccdtri:
				ccdtri[tri1] = ccdata([vts[v] for v in tri1])
			if tri2 not in ccdtri:
				ccdtri[tri2] = ccdata([vts[v] for v in tri2])
			ins1 = ccinside(ccdtri[tri1], vts[ix2])
			ins2 = ccinside(ccdtri[tri2], vts[ix1])
			if ins1 or ins2:
				# Compose the new triangles and check them
				k1 = triixs[tri1]
				k2 = triixs[tri2]
				tri1 = tuple(RotateToMin((ed[0],ix2,ix1)))
				tri2 = tuple(RotateToMin((ed[1],ix1,ix2)))
				tor1 = orient([vts[v] for v in tri1])
				tor2 = orient([vts[v] for v in tri2])
				if tor1 and tor2:
					# Put in the new edge and triangles
					numflips += 1
					ed = tuple(sorted((ix1,ix2)))
					eds[k] = ed
					edixs[ed] = k
					tris[k1] = tri1
					tris[k2] = tri2
					triixs[tri1] = k1
					triixs[tri2] = k2
					TriEdgeRefs(tri1, trieds, edtri)
					TriEdgeRefs(tri2, trieds, edtri)
					
					# Mark all affected edges as to be examined
					for ed in trieds[tri1]:
						edtests[edixs[tuple(sorted(ed))]] = True
					for ed in trieds[tri2]:
						edtests[edixs[tuple(sorted(ed))]] = True
			else:
				# This edge does not need further examination in this step
				edtests[k] = False
		
		totflips.append(numflips)
		if numflips == 0: break
	
	return (tris, tuple(totflips))

def GenIncrFlipBase(vts, orient, datafunc, dataorient, ccdata, ccinside):
	trieds = {}
	edtri = {}
	hull, tris = \
		GenInitialTriangulation(vts, trieds, edtri, orient, datafunc, dataorient)
	tris, totflips = \
		GenEdgeFlipping(vts, tris, trieds, edtri, orient, ccdata, ccinside)
	
	return ((hull,), tuple(sorted(tris)), (), totflips)

def DT2ReindexRes(ixs, xlss):
	xlsres = [tuple(RotateToMin([ixs[x] for x in xlst])) for xlst in xlss]
	return tuple(sorted(xlsres))

def GenIncrFlip(vts, orient, datafunc, dataorient, ccdata, ccinside, sorttype):
	if sorttype == "lin":
		ixvts = AddIndices(vts)
		ixvts = [(iv[0], iv[1], iv[1][0]) for iv in ixvts]
		ixvts.sort(lambda a,b: cmp(a[2],b[2]))
		ixs = [iv[0] for iv in ixvts]
		svts = [iv[1] for iv in ixvts]
		vtssorted = True
	elif sorttype == "rad":
		ixvts = AddIndices(vts)
		avg = VectorMean(vts)
		ixvts = [(iv[0], iv[1], VectorAbsSqr(VectorSub(iv[1],avg))) for iv in ixvts]
		ixvts.sort(lambda a,b: cmp(a[2],b[2]))
		ixs = [iv[0] for iv in ixvts]
		svts = [iv[1] for iv in ixvts]
		vtssorted = True
	else:
		svts = vts
		vtssorted = False
		
	dtres = GenIncrFlipBase(svts, orient, datafunc, dataorient, ccdata, ccinside)
	if vtssorted:
		dtres = (DT2ReindexRes(ixs, dtres[0]), DT2ReindexRes(ixs, dtres[1]), \
			dtres[2], dtres[3])
	
	return dtres

def IncrFlip(vts, sorttype=None):
	return GenIncrFlip(vts, TriOrient2D, TriAreaData2D, TriOrientEval2D, \
		CircumData, CircumInside, sorttype)

def IncrFlipLin(vts): return IncrFlip(vts, "lin")

def IncrFlipRad(vts): return IncrFlip(vts, "rad")

def SphIncrFlip(vts, sorttype=None):
	return GenIncrFlip(vts, TriOrient2Sph, TriAreaData2Sph, TriOrientEval2Sph, \
		SphCircumData, SphCircumInside, sorttype)

def SphIncrFlipLin(vts): return SphIncrFlip(vts, "lin")

def SphIncrFlipRad(vts): return SphIncrFlip(vts, "rad")


# General - Convex Hull

# Split a hull using a ring. This will work for any simplex dimension
def SplitHull(faces, ringsubfcs):
	if len(faces) == 0: return ((), ())
	if len(ringsubfcs) == 0: return (tuple(faces), ())
	
	# Get the face and subface references, their references to each other
	fctosbfcs = {}
	sbfctofcs = {}
	for face in faces:
		SimplexSubRefs(face, fctosbfcs, sbfctofcs)
	
	# Ring edges: don't expand past those
	sbfcuse = {}
	for face in faces:
		for sbfc in fctosbfcs[face]:
			sbfcuse[sbfc] = True
	for sbfc in ringsubfcs:
		sbfcuse[sbfc] = False
	
	# Find which existing faces are accessible from the ring subfaces.
	# 0 = unchecked, 1 = checked
	
	rngfaces = [sbfctofcs[sbfc] for sbfc in ringsubfcs]
	facestat = {}
	for face in rngfaces:
		facestat[face] = 0
	unchecked = True
	
	while unchecked:
		unchecked = False
		for k, face in enumerate(rngfaces):
			if facestat[face] == 0:
				unchecked = True
				
				# Look at a triangle's neighboring triangles
				for sbfc in fctosbfcs[face]:
					if sbfcuse[sbfc]:
						facex = sbfctofcs[SimplexFlip(sbfc)]
						if facex not in facestat:
							rngfaces.append(facex)
							facestat[facex] = 0
				
				facestat[face] = 1
	
	rngfaceset = set(rngfaces)
	farfaceset = set(faces) - rngfaceset
	rngfaces = tuple(sorted(rngfaceset))
	farfaces = tuple(sorted(farfaceset))
	return (rngfaces, farfaces)

# Append the square to put the vertices on a parabola/paraboloid
# in a 1 higher dimensional space.
# Also make all the variation between -1 and 1
def ParabExtend(vts):
	vtsmean = VectorMean(vts)
	adjvts = [VectorSub(vt,vtsmean) for vt in vts]
	vtsnorm = max(map(VectorNorm,adjvts))
	if vtsnorm == 0: vtsnorm = 1
	normrecip = 1./vtsnorm
	for vt in adjvts:
		VectorScalMultBy(normrecip, vt)
	
	return [list(vt) + [VectorAbsSqr(vt)] for vt in adjvts]


def Identity(x): return x

# Which function to use
BoundaryCH2 = ConvexHull2D.QuickHull
BoundaryCH3 = ConvexHull3D.QuickHull
BoundaryCHN = ConvexHullND.QuickHull
SphBoundaryCH2 = ConvexHull2D.SphIncremental
SphBoundaryCHN = ConvexHullND.SphIncremental
TessCH3 = ConvexHull3D.QuickHull
TessCHN = ConvexHullND.QuickHull

# Returns the hull and the faces in the hull *)
def BoundaryCH(vts, ndims):
	if ndims == 2:
		hull = BoundaryCH2(vts)
		nh = len(hull)
		hullfaces = []
		for k in xrange(nh):
			kx = (k+1) % nh
			seg = (hull[k],hull[kx])
			hullfaces.append(seg)
		hull = (hull,)
		hullfaces = tuple(hullfaces)
	else:
		if ndims == 3:
			hull = BoundaryCH3(vts)
		else:
			hull = BoundaryCHN(vts)
		hullfaces = hull[1]
	
	return (hull, hullfaces)

def SphBoundaryCH(vts, ndims):
	if ndims == 2:
		hull = SphBoundaryCH2(vts)
		nh = len(hull)
		hullfaces = []
		for k in xrange(nh):
			kx = (k+1) % nh
			seg = (hull[k],hull[kx])
			hullfaces.append(seg)
		hull = (hull,)
		hullfaces = tuple(hullfaces)
	else:
		hull = SphBoundaryCHN(vts)
		hullfaces = hull[1]
	
	return (hull, hullfaces)
		
def TessCH(vts, ndims):
	if ndims == 3:
		hull = TessCH3(vts)
	else:
		hull = TessCHN(vts)
	hullfaces = hull[1]
	
	return (hull, hullfaces)

def FlipTessCH(vts, ndims):
	if ndims == 3:
		hull = TessCH3(vts)
	else:
		hull = TessCHN(vts)
	hull = (hull[0], tuple(sorted(map(SimplexFlip,hull[1]))))
	hullfaces = hull[1]
	
	return (hull, hullfaces)

def GenConvexHull(vts, extradims, GenBoundaryCH, ExtendVts, GenTessCH):
	nvts = len(vts)
	ndims = len(vts[0])
	ndims -= extradims
	if ndims < 2: return ((),(),())
	bdryhull, bdryhullfaces = GenBoundaryCH(vts, ndims)
	extvts = ExtendVts(vts)
	tesshull, tesshullfaces = GenTessCH(extvts, ndims+1)
	dtfaces, frfaces = SplitHull(tesshullfaces, bdryhullfaces)
	return (bdryhull, dtfaces, frfaces)


def ConvexHull(vts):
	return GenConvexHull(vts, 0, BoundaryCH, ParabExtend, TessCH)

def SphConvexHull(vts):
	return GenConvexHull(vts, 1, SphBoundaryCH, Identity, FlipTessCH)
